;;; org-install.el --- backward compatibility file for obsolete configuration  -*- lexical-binding: t -*-
;;
;;; Code:
;;
;; The file org-install is obsolete.
;;
;; It is provided here so that (require 'org-install) does not
;; trigger an error for users with obsolete Emacs configuration.
;; You can safely remove (require 'org-install) from your config."

(provide 'org-install)

;; Local Variables:
;; no-byte-compile: t
;; coding: utf-8
;; End:
;;; org-install.el ends here
